#!/usr/bin/env python
#
# PkgTrackInst.Trackers - Tracks installation log
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42, (c) Poul-Henning Kamp):
# Hye-Shik Chang <perky@FreeBSD.org> wrote this file.  As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some
# day, and you think this stuff is worth it, you can buy me a beer in return.
#
# Hye-Shik Chang
# ----------------------------------------------------------------------------
# $Id: Trackers.py 478 2005-05-01 14:42:01Z perky $

__pychecker__ = 'no-argsused no-isliteral'

import os
from PkgTrackInst import options
from PkgTrackInst.cmd_support import errx

def safe_oct(v):
    if v is None:
        return 'None'
    else:
        return oct(v)


class InstEntry(object):

    root = None             # singleton object for root directory
    abspath = None
    created = False
    mode = None
    owner = (None, None)
    realpath = None         # for InstSymbolicLink only

    def __init__(self, created=False):
        self.created = created
        self.owner = [None, None]

    def chmod(self, newmode, symlink=False):
        if isinstance(self, InstSymbolicLink) and not symlink:
            self.root[self.realpath].chmod(newmode, symlink)
        else:
            self.mode = newmode

    def chown(self, uid, gid, symlink=False):
        if isinstance(self, InstSymbolicLink) and not symlink:
            dst = self.root[self.realpath]
        else:
            dst = self

        if uid is not None and uid >= 0:
            dst.owner[0] = uid
        if gid is not None and gid >= 0:
            dst.owner[1] = gid

    @staticmethod
    def create_root():
        InstEntry.root = InstDirectory()
        InstEntry.root.abspath = '/'
        return InstEntry.root

    def traverse(self, depth=0):
        print '  ' * depth + repr(self)

    def statup(self):
        if self.abspath is not None and self.created:
            try:
                stat = os.lstat(self.abspath)
                if self.mode is None:
                    self.chmod(stat.st_mode & 03777)
                if self.owner == [None, None]:
                    self.chown(stat.st_uid, stat.st_gid)
            except:
                pass # ignore when deleted after all

    def iterchildren(self):
        return iter([])

    def walk(self, callback):
        callback(self)
        for name, child in self.iterchildren():
            child.walk(callback)


class InstDirectory(InstEntry):

    def __init__(self, created=False):
        InstEntry.__init__(self, created)
        self.entries = {}

    def __getitem__(self, name):
        pathparts = name.split(os.sep)
        if self.abspath == '/' and not pathparts[0]:
            del pathparts[0]
        if len(pathparts) > 1:
            return self[pathparts[0]][os.sep.join(pathparts[1:])]

        name = pathparts[0]
        if name not in self.entries:
            path = os.path.join(self.abspath, name)
            if os.path.isdir(path):
                self.entries[name] = InstDirectory()
                self.entries[name].abspath = path
            else: # untracked path
                raise KeyError, os.path.join(self.abspath, name)
        return self.entries[name]

    def __setitem__(self, path, obj):
        assert isinstance(obj, InstEntry)

        pathparts = self.split_path(path)
        if len(pathparts) == 1:
            self.entries[pathparts[0]] = obj
            obj.abspath = os.path.join(self.abspath, pathparts[0])
            obj.statup()
        else:
            self[pathparts[0]][os.sep.join(pathparts[1:])] = obj

    def __delitem__(self, path):
        pathparts = self.split_path(path)
        if len(pathparts) == 1:
            if pathparts[0] not in self.entries:
                return # untracked object
            del self.entries[pathparts[0]]
        else:
            del self[pathparts[0]][os.sep.join(pathparts[1:])]

    def __repr__(self):
        return ('<InstDirectory abspath=%s created=%s mode=%s owner=%s '
                'children=%s>' % (
                    repr(self.abspath), repr(self.created), safe_oct(self.mode),
                    repr(self.owner), repr(self.entries.keys())))

    def move(self, srcfile, dst):
        assert len(srcfile.split(os.sep)) == 1
        if srcfile not in self.entries:
            # treat it like a new file.
            self.root[dst] = InstFile(created=True)
            return

        sobj = self.entries[srcfile]
        del self.entries[srcfile]
        self.root[dst] = sobj

        def fixup_abspath(newpath, obj):
            obj.abspath = newpath
            obj.statup()
            for name, child in obj.iterchildren():
                fixup_abspath(os.path.join(newpath, name), child)
        fixup_abspath(dst, sobj)

    def traverse(self, depth=0):
        print '  ' * depth + repr(self)
        for name, obj in self.entries.iteritems():
            print '  ' * depth, name, ':'
            obj.traverse(depth+1)

    def split_path(self, path):
        parts = path.split(os.sep)
        if self.abspath == '/' and not parts[0]:
            del parts[0]
        return parts

    def iterchildren(self):
        return self.entries.iteritems()


class InstFile(InstEntry):

    def __repr__(self):
        return '<InstFile created=%s mode=%s owner=%s>' % (
                    repr(self.created), safe_oct(self.mode), self.owner)


class InstSymbolicLink(InstEntry):

    def __init__(self, realpath, created=False):
        InstEntry.__init__(self, created)
        self.realpath = realpath

    def __repr__(self):
        return '<InstSymbolicLink created=%s realpath=%s mode=%s owner=%s>' % (
                    repr(self.created), repr(self.realpath),
                    safe_oct(self.mode), repr(self.owner))


# singleton for the root directory
TrackRoot = InstEntry.create_root()
